"""
PyDiscrete Subpackage
=====================

Provides:
- `Group`: main class interfacing with GAP to compute group information for particle model-building
"""

from .PyDiscrete import Group

__all__ = ["Group"]
__version__ = "0.2.0"
